/****************************************************************************
 *
 * reduction.c - Tree-structured reduction
 *
 * Last updated in 2019 by Moreno Marzolla <moreno.marzolla(at)unibo.it>
 *
 * To the extent possible under law, the author(s) have dedicated all 
 * copyright and related and neighboring rights to this software to the 
 * public domain worldwide. This software is distributed without any warranty.
 *
 * You should have received a copy of the CC0 Public Domain Dedication
 * along with this software. If not, see 
 * <http://creativecommons.org/publicdomain/zero/1.0/>. 
 *
 * --------------------------------------------------------------------------
 *
 * Compile with:
 * gcc -std=c99 -Wall -pedantic reduction.c -o reduction
 *
 * Run with:
 * ./reduction
 *
 ****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <assert.h>

/* Serial implementation of the sum-reduce operator */
float sum_reduction_serial( float* x, size_t n )
{
    float result = 0.0;
    int i;
    for (i=0; i<n; i++) {
        result += x[i];
    }
    return result;
}

/* Tree-structured implementation of the sum-reduce operator; this
   implementation works for any vector length n > 0. Warning: this
   function modifies vector x!! */
float sum_reduction( float* x, size_t n )
{
    int i, gap = 2;

    /* Moving from the leaves to the root of the reduction tree; at
       each iteration we double the gap */
    for (gap = 1; gap < n; gap *= 2) {
        for (i=0; i<n; i += 2*gap) {
            /* The inner loop iterates over the nodes at the same
               level of the reduction tree. Processor (i) receives a
               value from processor (i+gap), provided that node
               (i+gap) exists.

               Processors (0, 2*gap, 4*gap, 6*gap...) receive;
               processors (gap, 3*gap, 5*gap, ...) transmit. */
            if (i+gap<n) {
                x[i] += x[i+gap];
            }
        }
    }
    return x[0];
}

/* Array length; this program should work for any value of N */
#define N 51

int main( void )
{
    float v[N], result, expect;
    int i;
    for (i=0; i<N; i++) {
	v[i] = i+1;
    }
    result = sum_reduction(v, N);
    expect = N*(N+1)/2.0;
    if ( fabs(result - expect) > 1e-5 ) {
        fprintf(stderr, "Error: expected %f, got %f\n", expect, result);
        return EXIT_FAILURE;
    }
    printf("Test ok\n");
    return EXIT_SUCCESS;
}
